#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "LzmaKWrapper.h"

#define P_UNKNOWN 0
#define P_SQUASHFS 1
#define P_LZMA 2
#define P_KERNELIMG 3

#define eqstr(a,b) ((a)==(b)?"OK":"!!")

char *ptypes[] = { "unknown", "squashfs", "lzma", "kernel-img" };

struct b {
	unsigned char *buf;
	int l;
};

struct sec {
	unsigned start;
	unsigned len;
	unsigned long long nothing;
};

struct mtd_hdr {
	char magic[4];
	unsigned prodid;
	unsigned long long relid;
	unsigned start;
	unsigned len;
	char infostr[0x40];
	unsigned hdrchk;
	unsigned imgchk;
	struct sec secs[];
};

struct lzma_khdr {
	char magic[6];
	unsigned char major; /* I don't think they really are */
	unsigned char minor; /* but fuck knows what else they could be */
	unsigned inlen;
	unsigned inchk;
	unsigned dunno;
	unsigned outlen;
	unsigned outchk;
	unsigned searchme;
};

struct b *readf(char *fn, struct b *ret) {
	FILE *fp;
	
	if ((fp = fopen(fn, "r")) == NULL) {
		fprintf(stderr, "%s: cannot open\n", fn);
		return(NULL);
	}
	fseek(fp, 0, SEEK_END);
	ret->l = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	ret->buf = malloc(ret->l);
	fread(ret->buf, 1, ret->l, fp);
	fclose(fp);
	return(ret);
}

int writef(char *fn, struct b *dat) {
	FILE *fp;
	int ret;
	
	if ((fp = fopen(fn, "w")) == NULL) {
		fprintf(stderr, "%s: cannot open\n", fn);
		return(-1);
	}
	ret = fwrite(dat->buf, 1, dat->l, fp);
	fclose(fp);
	return(ret);
}

unsigned chksum32(unsigned char *buf, int len) {
	unsigned ret = 0;
	int x;
	
	for (x = 0; x < len; x++) ret += buf[x];
	return(ret);
}
#define chksum8(b,l) ((unsigned char)(chksum32((b),(l))&0x000000ff))

void print16(char *s) {
	int x;
	
	for (x = 0; x < 16; x++) printf("%02.2hhx ", s[x]);
	printf("\n");
}

int get_part(struct b *dat, struct b *ret) {
	static struct b *olddat;
	struct mtd_hdr *hdr;
	struct lzma_khdr *khdr;
	static struct sec *s = NULL;
	static int seq = 0;
	int type;
	int x;
	unsigned n;
	unsigned char *decomp;
	
	if (!dat) {
		dat = olddat;
	} else {
		s = NULL;
		seq = 0;
		olddat = dat;
	}
	hdr = (struct mtd_hdr *)dat->buf;
	if (!s) s = hdr->secs;
	type = -1;
	while (((void *)s - (void *)hdr) < hdr->start) {
		if ((!s->start) && (!s->len)) {
			s++;
			continue;
		} else {
			printf("Block %d: start %06.6x, length %06.6x\n", seq, s->start, s->len);
			ret->buf = dat->buf + s->start;
			ret->l = s->len;
			if (!strncmp(ret->buf, "hsqs", 4)) {
				printf("Found squashfs partition at block %d, length %06.6x\n", seq, ret->l);
				type = P_SQUASHFS;
			} else if (khdr = memmem(ret->buf, ret->l, "\x37\x7a\xbc\xaf\x27\x1c", 6)) {
				x = (void *)khdr - (void *)ret->buf;
				ret->l -= x;
				printf("Partition at block %d has LZMA image at offset %06.6x\n", seq, x);
				type = P_LZMA;
				ret->buf = (unsigned char *)khdr;
				n = chksum32(ret->buf + 0x20, khdr->inlen);
				printf("	\"Major\":	%02.2hhx\n	\"Minor\":	%02.2hhx\n	Data len:	%08.8x\n	Uncomp len:	%08.8x\n"
					   "	Data chk:	got %08.8x, calc %08.8x (%s)\n	Uncomp chk:	got %08.8x, calc ",
					   khdr->major, khdr->minor, khdr->inlen, khdr->outlen, khdr->inchk, n, eqstr(khdr->inchk, n), khdr->outchk);
				n = khdr->outlen;
				decomp = malloc(n);
				if (lzma_inflate(ret->buf, khdr->inlen, decomp, &n)) {
					printf("(Decompression failed, will write compressed data)\n");
					free(decomp);
				} else {
					n = chksum32(decomp, khdr->outlen);
					printf("%08.8x (%s)\n", n, eqstr(khdr->outchk, n));
					ret->buf = decomp;
					ret->l = khdr->outlen;
					type = P_KERNELIMG;
				}				
			} else {
				printf("Block %d has unknown partition type, starts: ", seq);
				print16(ret->buf);
				type = P_UNKNOWN;
			}
			s++;
			seq++;
			break;
		}
	}
	return(type);
}

int list_parts(struct b *dat, char *fn) {
	struct b part;
	struct mtd_hdr *hdr;
	int type;
	unsigned n, h;
	char *fn2, *s;
	
	hdr = (struct mtd_hdr *)dat->buf;
	if (strncmp(hdr->magic, "4DTM", 4)) {
		printf("%s: Not an MTD partition, starts: ", fn);
		print16(dat->buf);
		return(0);
	}
	h = hdr->start + hdr->len;
	if (h > dat->l) {
		printf("%s: Truncated partition. Cannot extract. (The real one should have the same start address and a length of at least %d bytes.)\n", 
			   fn, h);
		return(0);
	}
	printf("Dumping from file %s\n", fn);
	fn2 = alloca(strlen(fn) + 32);
	strcpy(fn2, fn);
	if ((s = strstr(fn2, ".bin")) && (!s[4])) *s = 0;
	s = &(fn2[strlen(fn2)]);
	printf(" Prod ID:	%08.8x\n Rel ID:	%016.16llx\n Data start:	%08.8x\n Data length:	%08.8x\n Info string:	%s\n",
		   hdr->prodid, hdr->relid, hdr->start, hdr->len, hdr->infostr);
	h = hdr->hdrchk;
	hdr->hdrchk = 0;
	n = chksum32(dat->buf, hdr->start);
	hdr->hdrchk = h;
	printf(" Hdr chk:	got %08.8x, calc %08.8x (%s)\n", hdr->hdrchk, n, eqstr(hdr->hdrchk, n));
	n = chksum32(dat->buf + hdr->start, hdr->len);
	printf(" Data chk:	got %08.8x, calc %08.8x (%s)\n", hdr->imgchk, n, eqstr(hdr->imgchk, n));
	n = 0;
	h = 1;
	while ((type = get_part((h ? dat : NULL), &part)) >= 0) {
		sprintf(s, "-%d.%s.bin", n, ptypes[type]);
		printf("Writing file %s, type %d, length %x\n", fn2, type, part.l);
		writef(fn2, &part);
		if (type == P_KERNELIMG) free(part.buf);
		n++;
		h = 0;
	}
	return(n);	
}

void lzinit(void) {
	unsigned char *b;
	int s;
	
	s = lzma_workspace_size();
	b = malloc(s);
	lzma_init(b, s);
}

int main(int argc, char *argv[]) {
	struct b dat;
	int x, n;

	lzinit();
	for (x = 1; x < argc; x++) {
		if (readf(argv[x], &dat)) {
			n = list_parts(&dat, argv[x]);
			printf("Extracted %d partitions from %s\n", n, argv[x]);
			free(dat.buf);
		}
	}
}
